#include <gl/glew.h>
#include <gl/freeglut.h>

#include <cuda_runtime.h>
#include <cuda_gl_interop.h>

#include "Metaball.h"
#include "BasicDef.h"

// lʂ߂ZxlD
float iso_value = 1.0f;

// ^{[@Œ`ꂽϐD
extern uint num_voxels;

// }[`OL[u@ňꎞIɎQƂϐD
uint max_num_vertices;
uint num_occupied_voxels;
uint total_num_vertices;

uint *d_num_vertices_table;
uint *d_triangle_table;
uint *d_voxel_vertices;
uint *d_voxel_vertices_scanned;
uint *d_voxel_occupied;
uint *d_voxel_occupied_scanned;
uint *d_compacted_voxel;

GLuint poly_pos_vbo, poly_normal_vbo;
struct cudaGraphicsResource *poly_pos_vbo_res, *poly_normal_vbo_res;
float4 *d_poly_pos, *d_poly_normal;

// O`̊֐D
extern void thrustScan(unsigned int *input, unsigned int num_elements, unsigned int *output);
extern void launchClassifyVoxel(uint num_voxels, float iso_value, uint *voxel_vertices, uint *voxel_occupied);
extern void launchCompactVoxels(uint num_voxels, uint *voxel_occupied, uint *voxel_occupied_scanned, uint *compacted_voxel);
extern void launchGeneratePolys(uint num_occupied_voxels, uint max_num_vertices, float iso_value, 
	uint *compacted_voxel, uint *num_vertices_scanned, float4 *poly_pos, float4 *poly_normal);

// _obt@IuWFNg̐D
void createVBO(GLuint *vbo, unsigned int size, struct cudaGraphicsResource **vbo_res, unsigned int vbo_res_flags)
// GLuint *vbo; _obt@IuWFNgD
// unsigned int size; obt@̃TCYD
// struct cudaGraphicsResource **vbo_res; CUDÃOtBbNX\[XD
// unsigned int vbo_res_flags; OtBbNX\[X̎gɂẴqgD
{
    glGenBuffers(1, vbo);
    glBindBuffer(GL_ARRAY_BUFFER, *vbo);
	glBufferData(GL_ARRAY_BUFFER, size, NULL, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // _obt@IuWFNgOtBbNX\[Xɓo^D
	cudaGraphicsGLRegisterBuffer(vbo_res, *vbo, vbo_res_flags);
}

// _obt@IuWFNg̍폜D
void deleteVBO(GLuint *vbo, struct cudaGraphicsResource *vbo_res)
// GLuint *vbo; _obt@IuWFNgD
// struct cudaGraphicsResource *vbo_res; CUDÃOtBbNX\[XD
{

	// _obt@IuWFNgo^OD
	cudaGraphicsUnregisterResource(vbo_res);
    glDeleteBuffers(1, vbo);
    *vbo = 0;
}

// }[`OL[u@ɂ铙lʂ̃|SD
void marchingCubesIsoSurface(void)
{
    uint last_elem, last_scan_elem;
 
    // |S\tZ̑ID
    launchClassifyVoxel(num_voxels, iso_value, d_voxel_vertices, d_voxel_occupied);

    // ތ̃ZXLrIɐώZD
	thrustScan(d_voxel_occupied, num_voxels, d_voxel_occupied_scanned);

    // XLʂɊÂďΏۂ̃Z𓾂DXLʂ
    // ̌̃Z̈kłpDrIȐώZȂ̂ŁCŌ̌
    // ʓrKvD
    cudaMemcpy((void *)&last_elem, (void *)(d_voxel_occupied + num_voxels - 1), 
        sizeof(uint), cudaMemcpyDeviceToHost);
    cudaMemcpy((void *)&last_scan_elem, (void *)(d_voxel_occupied_scanned + num_voxels - 1), 
        sizeof(uint), cudaMemcpyDeviceToHost);
    num_occupied_voxels = last_elem + last_scan_elem;

    // ׂZ[̃P[XD
    if (num_occupied_voxels == 0) {
        total_num_vertices = 0;
        return;
    }

    // Ώۂ̃Z̈kD
    launchCompactVoxels(num_voxels, d_voxel_occupied, d_voxel_occupied_scanned, d_compacted_voxel);

    // eZɂă|S\tɕKvȒ_rIɐώZD
	thrustScan(d_voxel_vertices, num_voxels, d_voxel_vertices_scanned);

    // Kvȑ_𓾂D
    cudaMemcpy((void *)&last_elem, (void *)(d_voxel_vertices + num_voxels - 1), 
        sizeof(uint), cudaMemcpyDeviceToHost);
    cudaMemcpy((void *)&last_scan_elem, (void *)(d_voxel_vertices_scanned + num_voxels - 1),
        sizeof(uint), cudaMemcpyDeviceToHost);
    total_num_vertices = last_elem + last_scan_elem;

    // _Ɩ@̃OtBbNX\[XfoCXϐd_poly_posd_poly_normalփ}bvD
	cudaGraphicsMapResources(1, &poly_pos_vbo_res, 0);
	cudaGraphicsResourceGetMappedPointer((void**)&d_poly_pos, NULL, poly_pos_vbo_res);
	cudaGraphicsMapResources(1, &poly_normal_vbo_res, 0);
	cudaGraphicsResourceGetMappedPointer((void**)&d_poly_normal, NULL, poly_normal_vbo_res);

	// Zւ̃|S̓\tD
    launchGeneratePolys(num_occupied_voxels, max_num_vertices, iso_value, d_compacted_voxel, d_voxel_vertices_scanned, 
        d_poly_pos, d_poly_normal);

	// A}bvD
	cudaGraphicsUnmapResources(1, &poly_pos_vbo_res, 0);
	cudaGraphicsUnmapResources(1, &poly_normal_vbo_res, 0);
}

// lʂ\|SQVBOp`D
void drawIsoSurface(void)
{
	glBindBuffer(GL_ARRAY_BUFFER, poly_pos_vbo);
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(4, GL_FLOAT, 0, 0);
	
	glBindBuffer(GL_ARRAY_BUFFER, poly_normal_vbo);
	glEnableClientState(GL_NORMAL_ARRAY);
	glNormalPointer(GL_FLOAT, 4 * sizeof(float), 0);

	glDrawArrays(GL_TRIANGLES, 0, total_num_vertices);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

